#!/usr/bin/env python
from __future__ import print_function
from bcc import BPF
from time import sleep
import argparse
import threading
#lock = threading.Lock()

bpf_text = """
#include <uapi/linux/ptrace.h>
#include <linux/sched.h>
struct key_t {
    u32 tgid;
    int stack_id;
    char name[TASK_COMM_LEN];
};
BPF_HASH(sizehash, int, u64);
BPF_HASH(counts, struct key_t);
BPF_STACK_TRACE(stack_traces, 1024);
BPF_HASH(sizes, u64 ,u64);
BPF_HASH(size_rea, int ,u64);

int count_malloc(struct pt_regs *ctx) {
    struct key_t key = {};
    key.tgid = bpf_get_current_pid_tgid() >> 32;
    bpf_get_current_comm(&key.name, sizeof(key.name));
    key.stack_id = stack_traces.get_stackid(
          ctx, BPF_F_USER_STACK);
    u64 zero = 0, *val;
    u64 size = ctx->di;
    val = counts.lookup_or_init(&key, &zero);
    (*val) += size;
    int i = bpf_get_current_pid_tgid() >> 32;
    sizehash.update(&i,&size);
    return 0;
}

int count_free(struct pt_regs *ctx) {
    struct key_t key = {};
    key.tgid = bpf_get_current_pid_tgid() >> 32;
    bpf_get_current_comm(&key.name, sizeof(key.name));
    key.stack_id = stack_traces.get_stackid(
          ctx, BPF_F_USER_STACK);
    u64 zero = 0, *val;
    val = counts.lookup_or_init(&key, &zero);
    u64 ptr = ctx->di;
    u64* size_ptr = sizes.lookup(&ptr);
    if(size_ptr){
        u64 size = *size_ptr;
        (*val) += size;
        sizes.delete(&ptr);
    }
    
    return 0;
}

int count_ret(struct pt_regs *ctx) {
    int i = bpf_get_current_pid_tgid() >> 32;
    u64 ptr  = ctx->ax;
    u64* size_ptr = sizehash.lookup(&i);
    if(size_ptr){
        u64 size = *size_ptr;
        sizes.update(&ptr,&size);
    }
    return 0;
}

int count_calloc(struct pt_regs *ctx) {
    struct key_t key = {};
    key.tgid = bpf_get_current_pid_tgid() >> 32;
    bpf_get_current_comm(&key.name, sizeof(key.name));
    key.stack_id = stack_traces.get_stackid(
          ctx, BPF_F_USER_STACK);
    u64 zero = 0, *val;
    u64 size1 = ctx->di;
    u64 size2 = ctx->si;
    u64 size = size1*size2;
    val = counts.lookup_or_init(&key, &zero);
    (*val) += size;
    int i = bpf_get_current_pid_tgid() >> 32;
    sizehash.update(&i,&size);
    return 0;
}

int count_realloc(struct pt_regs *ctx) {
    struct key_t key = {};
    key.tgid = bpf_get_current_pid_tgid() >> 32;
    bpf_get_current_comm(&key.name, sizeof(key.name));
    key.stack_id = stack_traces.get_stackid(
          ctx, BPF_F_USER_STACK);
    u64 zero = 0, *val;
    u64 size = ctx->si;
    u64 ptr  = ctx->di;
    val = counts.lookup_or_init(&key, &zero);
    u64* oldsize = sizes.lookup(&ptr);
    //(*val) += (size-*(oldsize));
    (*val) += size;
    sizes.update(&ptr,&zero);  //清0

    int i = bpf_get_current_pid_tgid() >> 32;
    size_rea.update(&i,&size);

    return 0;
}


int count_realloc_ret(struct pt_regs *ctx) {
    struct key_t key = {};
    key.tgid = bpf_get_current_pid_tgid() >> 32;
    bpf_get_current_comm(&key.name, sizeof(key.name));
    key.stack_id = stack_traces.get_stackid(
          ctx, BPF_F_USER_STACK);
    u64 zero = 0, *val;
    int i = bpf_get_current_pid_tgid() >> 32;
    u64* size_ptr = size_rea.lookup(&i);
    if(size_ptr){
        u64 size = *size_ptr;
        u64 ptr  = ctx->ax;
        sizes.update(&ptr,&size);
        val = counts.lookup_or_init(&key, &zero);
    }
    
    
    return 0;
}

        """

parser = argparse.ArgumentParser()
parser.add_argument("-p", "--pid", type=int)
args = parser.parse_args()
exiting = 0

def count():

    #lock.acquire()
    print(" Hit Ctrl-C to end.")
    print()
    b = BPF(text=bpf_text)
    #lock.release()
    library = BPF.find_library("c")
    b.attach_uprobe(name=library,sym_re="malloc",fn_name="count_malloc",pid=args.pid)
    b.attach_uretprobe(name=library,sym_re="malloc",fn_name="count_ret",pid=args.pid)

    b.attach_uprobe(name=library,sym_re="calloc",fn_name="count_calloc",pid=args.pid)
    b.attach_uretprobe(name=library,sym_re="calloc",fn_name="count_ret",pid=args.pid)
    
    b.attach_uprobe(name=library,sym_re="realloc",fn_name="count_realloc",pid=args.pid)
    b.attach_uretprobe(name=library,sym_re="realloc",fn_name="count_realloc_ret",pid=args.pid)
    
    b.attach_uprobe(name=library,sym_re="free",fn_name="count_free",pid=args.pid)
    
    while exiting == 0:
        sleep(1)

    counts = b["counts"]
    stack_traces = b["stack_traces"]
    for k, v in sorted(counts.items(),key=lambda counts: counts[1].value):
        stack = list(stack_traces.walk(k.stack_id))
        print()
        for addr in stack:
            print("  ", end="")
            print("%s" % b.sym(addr, k.tgid).decode())
        print("  %d\n" % v.value)
    counts.clear()
    exit()

thread1 = threading.Thread(target=count)
thread1.start()
try:
    sleep(1000000)
except KeyboardInterrupt:
    exiting = 1